@@ -1,7 +1,7 @@
 ;;; -*- Mode: LISP; Syntax: COMMON-LISP; Package:CL-WHO; Base: 10 -*-
-;;; $Header: /usr/local/cvsrep/cl-who/who.lisp,v 1.27 2006/10/10 07:38:02 edi Exp $
+;;; $Header: /usr/local/cvsrep/cl-who/who.lisp,v 1.33 2007/05/28 18:31:31 edi Exp $
 
-;;; Copyright (c) 2003-2006, Dr. Edmund Weitz. All rights reserved.
+;;; Copyright (c) 2003-2007, Dr. Edmund Weitz. All rights reserved.
 
 ;;; Redistribution and use in source and binary forms, with or without
 ;;; modification, are permitted provided that the following conditions
@@ -27,42 +27,7 @@
 ;;; NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 ;;; SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-(in-package #:cl-who)
-
-(defvar *prologue*
-  "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">"
-  "This is the first line that'll be printed if the :PROLOGUE keyword
-argument is T")
-
-(defparameter *escape-char-p*
-  #'(lambda (char)
-      (or (find char "<>&'\"")
-          (> (char-code char) 127)))
-  "Used by ESCAPE-STRING to test whether a character should be escaped.")
-
-(defparameter *indent* nil
-  "Whether to insert line breaks and indent. Also controls amount of
-indentation dynamically.")
-
-(defvar *html-mode* :xml
-  ":SGML for \(SGML-)HTML, :XML \(default) for XHTML.")
-
-(defparameter *attribute-quote-char* #\'
-  "Quote character for attributes.")
-
-(defparameter *empty-tag-end* " />"
-  "End of an empty tag. Default is XML style.")
-
-(eval-when (:compile-toplevel :load-toplevel :execute)
-  (unless (boundp '+newline+)
-    (defconstant +newline+ (make-string 1 :initial-element #\Newline)
-      "Used for indentation."))
-
-  (unless (boundp '+spaces+)
-    (defconstant +spaces+ (make-string 2000
-                                       :initial-element #\Space
-                                       :element-type 'base-char)
-      "Used for indentation.")))
+(in-package :cl-who)
 
 (defmacro n-spaces (n)
   "A string with N spaces - used by indentation."
@@ -219,15 +184,18 @@ forms."
             ;; do the same things as above but at runtime
             nconc (list `(let ((,=var= ,val))
                           (cond ((null ,=var=))
-                                ((and (eq ,=var= t) (eq *html-mode* :xml))
-                                 (htm ,(format nil " ~A=~C~A~C"
-                                               (string-downcase attr)
-                                               *attribute-quote-char*
-                                               (string-downcase attr)
-                                               *attribute-quote-char*)))
-                                ((and (eq ,=var= t) (eq *html-mode* :sgml))
-                                 (htm ,(format nil " ~A"
-                                               (string-downcase attr))))
+                                ((eq ,=var= t) 
+                                 ,(case *html-mode*
+                                        (:sgml
+                                         `(htm ,(format nil " ~A"
+                                                        (string-downcase attr))))
+                                        ;; otherwise default to :xml mode
+                                        (t
+                                         `(htm ,(format nil " ~A=~C~A~C"
+                                                        (string-downcase attr)
+                                                        *attribute-quote-char*
+                                                        (string-downcase attr)
+                                                        *attribute-quote-char*)))))
                                 (t
                                  (htm ,(format nil " ~A=~C" (string-downcase attr)
                                                *attribute-quote-char*)
@@ -246,12 +214,13 @@ a list of strings or Lisp forms."))
   "The standard method which is not specialized.  The idea is that you
 can use EQL specializers on the first argument."
   (declare (optimize speed space))
+  (let ((tag (if *downcase-tags-p* (string-downcase tag) (string tag))))
     (nconc
      (if *indent*
        ;; indent by *INDENT* spaces
        (list +newline+ (n-spaces *indent*)))
      ;; tag name
-     (list "<" (string-downcase tag))
+     (list "<" tag)
      ;; attributes
      (convert-attributes attr-list)
      ;; body
@@ -268,9 +237,12 @@ can use EQL specializers on the first argument."
           ;; indentation
           (list +newline+ (n-spaces *indent*)))
         ;; closing tag
-        (list "</" (string-downcase tag) ">"))
-       ;; no body, so no closing tag
-       (list *empty-tag-end*))))
+        (list "</" tag ">"))
+       ;; no body, so no closing tag unless defined in *HTML-EMPTY-TAGS*
+       (if (or (not *html-empty-tag-aware-p*)
+               (member tag *html-empty-tags* :test #'string-equal))
+         (list *empty-tag-end*)
+         (list ">" "</" tag ">"))))))
 
 (defun apply-to-tree (function test tree)
   (declare (optimize speed space))
@@ -400,19 +372,23 @@ for REPLACE-HTM."
   (apply-to-tree #'(lambda (x)
                      (case (first x)
                        ((esc)
-                         ;; (ESC form ...) ->
-                         ;; (WRITE-STRING (ESCAPE-STRING form STREAM))
-                         (list 'write-string
-                               (list 'escape-string
-                                     (second x))
-                               stream))
+                        ;; (ESC form ...)
+                        ;; --> (LET ((RESULT form))
+                        ;;       (WHEN RESULT
+                        ;;         (WRITE-STRING (ESCAPE-STRING RESULT STREAM))))
+                        (let ((result (gensym)))
+                          `(let ((,result ,(second x)))
+                             (when ,result (write-string (escape-string ,result) ,stream)))))
                        ((str)
-                         ;; (STR form ...) --> (PRINC form STREAM)
-                         `(princ ,(second x) ,stream))
+                        ;; (STR form ...)
+                        ;; --> (LET ((RESULT form))
+                        ;;       (WHEN RESULT (PRINC RESULT STREAM)))
+                        (let ((result (gensym)))
+                          `(let ((,result ,(second x)))
+                             (when ,result (princ ,result ,stream)))))
                        ((fmt)
-                         ;; (FMT form*) --> (FORMAT STREAM form*)
-                         (cons 'format
-                               (cons stream (rest x))))))
+                        ;; (FMT form*) --> (FORMAT STREAM form*)
+                        (list* 'format stream (rest x)))))
                  #'(lambda (x)
                      (and (consp x)
                           (member (first x)
@@ -424,9 +400,10 @@ for REPLACE-HTM."
                                     (and (atom x)
                                          (eq x 'htm)))
                                 (tree-to-commands-aux
-                                 (cons 'htm
-                                       (cons prologue
-                                             (tree-to-template tree)))
+                                 (if prologue
+                                   (list* 'htm prologue +newline+
+                                          (tree-to-template tree))
+                                   (cons 'htm (tree-to-template tree)))
                                  stream))))
 
 (defmacro with-html-output ((var &optional stream
